
"""
ID list class for Jaseci

Generalized functions for managing '_ids' convention for lists of Jaseci
objects

parent_obj is the instance that the list belongs to
"""
from jaseci.utils.utils import logger
import uuid


class id_list(list):
    """
    ID list class for tracking lists of objects in Jaseci

    ingest_list is a list of hex strings to convert to UUID and append.
    """

    def __init__(self, parent_obj, in_list=None):
        self.parent_obj = parent_obj
        if (in_list):
            for i in in_list:
                self.append(i)

    def add_obj(self, obj, allow_dups=False, silent=False):
        """Adds a obj obj to Jaseci object"""
        self.parent_obj.check_hooks_match(obj)
        if not allow_dups and obj.jid in self:
            if(not silent):
                logger.warning(
                    str(f"{obj} is already in {self.parent_obj}'s list")
                )
        else:
            self.append(obj.jid)
            if(not obj.parent_id):
                obj.parent_id = self.parent_obj.id
            obj.save()
            self.parent_obj.save()

    def add_obj_list(self, obj_list, allow_dups=False, silent=False):
        for i in obj_list:
            self.add_obj(i, allow_dups=allow_dups, silent=silent)

    def remove_obj(self, obj):
        """Remove a Jaseci obj from list"""
        self.remove(obj.jid)
        self.parent_obj.save()

    def destroy_obj(self, obj):
        """Completely destroys a Jaseci obj obj by it's name"""
        self.remove_obj(obj)
        obj.destroy()

    def get_obj_by_name(self, name, kind=None, silent=False):
        """Returns a Jaseci obj obj by it's name"""
        healing = []
        ret = None
        for i in self:
            if (not self.parent_obj._h.has_obj(uuid.UUID(i))):
                logger.critical(str(
                    f'Self healing: {i} not found ' +
                    f'in id_list of {self.parent_obj}!'))
                healing.append(i)
                continue
            obj = self.parent_obj._h.get_obj(
                self.parent_obj._m_id, uuid.UUID(i))
            if(obj.name == name):
                if(kind and obj.kind != kind):
                    continue
                ret = self.parent_obj._h.get_obj(
                    self.parent_obj._m_id, uuid.UUID(i))
                break
        self.heal(healing)
        if (not ret and not silent):
            logger.error(
                str(f"object for '{name}' not found in '{self.parent_obj}'!")
            )
        return ret

    def has_obj_by_name(self, name, kind=None):
        """Returns whether a Jaseci obj exists by it's name"""
        return self.get_obj_by_name(
            name, kind, silent=True
        ) is not None

    def remove_obj_by_name(self, name, kind=None):
        """Remove a Jaseci obj by it's name"""
        self.remove_obj(self.get_obj_by_name(name, kind))

    def destroy_obj_by_name(self, name, kind=None):
        """Destroy a Jaseci obj by it's name"""
        self.destroy_obj(self.get_obj_by_name(name, kind))

    def obj_list(self):
        """Return list of objects from ids"""
        ret = []
        healing = []
        for i in self:
            obj = self.parent_obj._h.get_obj(
                self.parent_obj._m_id, uuid.UUID(i))
            if (not obj):
                logger.critical(str(
                    f'Self healing: {i} not found ' +
                    f'in id_list of {self.parent_obj}!'))
                healing.append(i)
            else:
                ret.append(self.parent_obj._h.get_obj(
                    self.parent_obj._m_id, uuid.UUID(i)))
        self.heal(healing)
        return ret

    def remove_all(self):
        """Remove a Jaseci obj obj by it's name"""
        for i in self.obj_list():
            self.remove_obj(i)
        if (len(self)):
            logger.critical(str(
                f'Removeall all failed in id_list of {self.parent_obj} - ' +
                f'still has {self}!'))

    def destroy_all(self):
        """Remove a Jaseci obj obj by it's name"""
        for i in self.obj_list():
            self.destroy_obj(i)
        if (len(self)):
            logger.critical(str(
                f'Destroy all failed in id_list of {self.parent_obj} - ' +
                f'still has {self}!'))

    def heal(self, healing):
        if(len(healing)):
            for i in healing:
                self.remove(i)
            self.parent_obj.save()

    def first_obj(self):
        """Get first object in list"""
        if(not self):
            logger.error(
                str(f"List in '{self.parent_obj}' is empty!")
            )
            return None
        return self.parent_obj._h.get_obj(
            self.parent_obj._m_id, uuid.UUID(self[0]))

    def pop_first_obj(self):
        """Get first object in list"""
        ret = self.first_obj()
        if(ret):
            self.remove_obj(ret)
        return ret



"""
Jaseci object mixins

Various mixins to define properties of Jaseci objects
"""
from jaseci.utils.id_list import id_list
from jaseci.utils.utils import logger
import uuid


class anchored():
    """Utility class for objects that hold anchor values"""

    def __init__(self):
        self.anchor = None

    def anchor_value(self):
        """Returns value of anchor context object"""
        if(self.anchor):
            return self.context[self.anchor]
        return None


class sharable():
    """Utility class for objects that are sharable between users"""

    def __init__(self, m_id, mode='private'):
        self.set_master(m_id)
        self.j_access = mode
        self.j_r_acc_ids = id_list(self)
        self.j_rw_acc_ids = id_list(self)

    @property
    def _m_id(self) -> str:
        return self.j_master

    def set_master(self, m_id):
        if(m_id is None or m_id == 'anon'):
            m_id = uuid.UUID(int=0).urn
        self.j_master = m_id

    def make_public(self):
        """Make element publically accessible"""
        self.j_access = 'public'
        self.save()

    def make_read_only(self):
        """Make element publically readable"""
        self.j_access = 'read_only'
        self.save()

    def make_private(self):
        """Make element private"""
        self.j_access = 'private'
        self.save()

    def is_public(self):
        """Check if element is publically accessible"""
        return self.j_access == 'public' or \
            self.j_master == uuid.UUID(int=0).urn

    def is_read_only(self):
        """Check if element is publically readable"""
        return self.j_access == 'read_only'

    def is_readable(self):
        """Check if element is publically readable"""
        return self.j_access == 'read_only' or self.is_public()

    def is_private(self):
        """Check if element is private"""
        return self.j_access == 'private'

    def check_read_access(self, caller_id, silent=False):
        if(caller_id == self._m_id or self.is_readable() or
           caller_id in self.j_r_acc_ids or
           caller_id in self.j_rw_acc_ids):
            return True
        if(not silent):
            logger.error(str(
                f'{caller_id} does not have permission to access {self}'))
        return False

    def check_write_access(self, caller_id, silent=False):
        if(caller_id == self._m_id or self.is_public() or
           caller_id in self.j_rw_acc_ids):
            return True
        if(not silent):
            logger.error(str(
                f'{caller_id} does not have permission to access {self}'))
        return False

    def give_access(self, m, read_only=True):
        """Give access to a master (user)"""
        if(not m.is_master()):
            logger.error(f'{m} is not master!')
            return False
        if(read_only and m.jid not in self.j_r_acc_ids):
            self.j_r_acc_ids.add_obj(m)
        elif(m.jid not in self.j_rw_acc_ids):
            self.j_rw_acc_ids.add_obj(m)
        return True

    def remove_access(self, m):
        """Remove access from a master (user)"""
        ret = False
        if(m.jid in self.j_r_acc_ids):
            self.j_r_acc_ids.remove_obj(m)
            ret = True
        if(m.jid in self.j_rw_acc_ids):
            self.j_rw_acc_ids.remove_obj(m)
            ret = True
        return ret


class hookable(sharable):
    """Utility class for objects that are savable to DBs and other stores"""

    def __init__(self, h, m_id, persist: bool = True, *args, **kwargs):
        self._h = h  # hook for storing and loading to persistent store
        self._persist = persist
        sharable.__init__(self, m_id, *args, **kwargs)

    def check_hooks_match(self, target, silent=False):
        """Checks whether target object hook matches self's hook"""
        if(not silent and target._h != self._h):
            logger.critical(str("Hook for {} does not match {}, {} != {}".
                                format(target, self, target._h, self._h)))
        return target._h == self._h

    def save(self):
        """
        Write self through hook to persistent storage
        """
        self._h.save_obj(self._m_id, self, self._persist)

    def destroy(self):
        """
        Destroys self from persistent storage

        Note that the object will still exist in python until GC'd
        """
        self._h.destroy_obj(self._m_id, self, self._persist)
        del self

    def parent(self):
        """
        Returns the objects for list of owners of this element
        """
        return self._h.get_obj(self._m_id, self.parent_id)



"""
General master interface engine for client interfaces as mixin
"""
import uuid
from inspect import signature, getdoc
from jaseci.utils.utils import logger
from jaseci.utils.utils import is_jsonable
from jaseci.element.element import element


class interface():
    """
    General master interface engine
    """
    _public_api = []
    _private_api = []
    _admin_api = []
    _cli_api = []

    def __init__(self):
        """
        self._pub_committer is set by api implementaiton if intent
        is to commit changes enacted by public call
        """
        self._pub_committer = None

    def assimilate_api(api_list, func, cmd_group=None,
                       cli_args=None):
        cmd_group = func.__name__.split(
            '_') if cmd_group is None else cmd_group
        api_list.append(
            {'fname': func.__name__, 'sig': signature(func),
             'doc': getdoc(func), 'groups': cmd_group,
             'cli_args': cli_args if cli_args is not None else []})
        return func

    def public_api(cmd_group=None, cli_args=None):
        def decorator_func(func):
            return interface.assimilate_api(
                interface._public_api, func, cmd_group, cli_args)
        return decorator_func

    def private_api(cmd_group=None, cli_args=None):
        def decorator_func(func):
            return interface.assimilate_api(
                interface._private_api, func, cmd_group, cli_args)
        return decorator_func

    def admin_api(cmd_group=None, cli_args=None):
        def decorator_func(func):
            return interface.assimilate_api(
                interface._admin_api, func, cmd_group, cli_args)
        return decorator_func

    def cli_api(cmd_group=None, cli_args=None):
        def decorator_func(func):
            return interface.assimilate_api(
                interface._cli_api, func, cmd_group, cli_args)
        return decorator_func

    def all_apis(self):
        return interface._public_api+interface._private_api + \
            interface._admin_api

    assimilate_api = staticmethod(assimilate_api)
    public_api = staticmethod(public_api)
    private_api = staticmethod(private_api)
    admin_api = staticmethod(admin_api)
    cli_api = staticmethod(cli_api)

    def provide_internal_default(self, param):
        """
        Applies internal defaults for sentinel and graphs
        """
        if(param == 'snt' and self.active_snt_id):
            if(self.active_snt_id == 'global'):
                glob_id = self._h.get_glob('GLOB_SENTINEL')
                if(not glob_id):
                    return self.interface_error(
                        'No global sentinel is available!')
                else:
                    return glob_id
            return self.active_snt_id
        if(param == 'gph' or param == 'nd'):
            return self.active_gph_id if self.active_gph_id else \
                self.interface_error('No default graph node available!')
        return None

    def interface_error(self, err, stack=None):
        """Standard error output to logger and api response"""
        logger.error(err)
        ret = {'response': err, 'success': False, 'errors': [err]}
        if(stack):
            ret['stack_trace'] = stack
        return ret

    def general_interface_to_api(self, params, api_name):
        """
        A mapper utility to interface to master class
        Assumptions:
            params is a dictionary of parameter names and values in UUID
            api_name is the name of the api being mapped to
        """
        param_map = {}
        if(api_name.startswith('master_active')):
            _caller = self
        else:
            _caller = self._caller
        if (not hasattr(_caller, api_name)):
            return self.interface_error(f'{api_name} not a valid API')
        func_sig = signature(getattr(_caller, api_name))
        for i in func_sig.parameters.keys():
            if (i == 'self'):
                continue
            p_name = i
            p_type = func_sig.parameters[i].annotation
            p_default = func_sig.parameters[i].default
            val = p_default if p_default is not \
                func_sig.parameters[i].empty else None
            if (p_name in params.keys()):
                val = params[p_name]
            if(val is None):  # Used to patch defaults
                val = _caller.provide_internal_default(p_name)
                if(val is not None and 'errors' in val):
                    return val
            if(str(val) in _caller.alias_map.keys()):
                val = _caller.alias_map[val]
            if (issubclass(p_type, element)):
                if(val is None):
                    break
                val = _caller._h.get_obj(
                    _caller._m_id, uuid.UUID(val))
                if (isinstance(val, p_type)):
                    param_map[i] = val
                else:
                    return self.interface_error(
                        f'{type(val)} is not {p_type}')
            else:  # TODO: Can do type checks here too
                param_map[i] = val

            if (param_map[i] is None):
                return self.interface_error(f'Invalid API args - {params}')
        try:
            ret = getattr(_caller, api_name)(**param_map)
        except Exception as e:
            import traceback as tb
            return self.interface_error(
                f'Internal Exception: {e}', stack=tb.format_exc())
        if(not is_jsonable(ret)):
            return self.interface_error(f'Non-JSON API ret {type(ret)}: {ret}')
        return ret

    def public_interface_to_api(self, params, api_name):
        """
        A mapper utility to interface to public
        Assumptions:
            params is a dictionary of parameter names and values in UUID
            api_name is the name of the api being mapped to
        """
        param_map = {}
        if (not hasattr(self, api_name)):
            return self.interface_error(f'{api_name} not a valid API')
            return False
        func_sig = signature(getattr(self, api_name))
        for i in func_sig.parameters.keys():
            if (i == 'self'):
                continue
            p_name = i
            p_type = func_sig.parameters[i].annotation
            p_default = func_sig.parameters[i].default
            val = p_default if p_default is not \
                func_sig.parameters[i].empty else None
            if (p_name in params.keys()):
                val = params[p_name]
            if (issubclass(p_type, element)):
                if(val is None):
                    return self.interface_error(
                        f'No {p_type} value for {p_name} provided!')
                val = self._h.get_obj(
                    'override', uuid.UUID(val), override=True)
                self.seek_committer(val)
                if (isinstance(val, p_type)):
                    param_map[i] = val
                else:
                    return self.interface_error(f'{type(val)} is not {p_type}')
                    param_map[i] = None
            else:  # TODO: Can do type checks here too
                param_map[i] = val

            if (param_map[i] is None):
                return self.interface_error(
                    f'Invalid API parameter set - {params}')
        try:
            ret = getattr(self, api_name)(**param_map)
        except Exception as e:
            import traceback as tb
            return self.interface_error(
                f'Internal Exception: {e}', stack=tb.format_exc())
        if(not is_jsonable(ret)):
            return self.interface_error(
                str(f'API returns non json object {type(ret)}: {ret}'))
        return ret

    def seek_committer(self, obj):
        """Opportunistically assign a committer"""
        if(not self._pub_committer):
            self._pub_committer = obj._h.get_obj(
                obj._m_id, uuid.UUID(obj._m_id))

    def clear_committer(self):
        """Unset committer"""
        self._pub_committer = None

    def sync_walker_from_global_sent(self, wlk):
        """Checks for matching code ir between global and spawned walker"""
        glob_id = wlk._h.get_glob('GLOB_SENTINEL')
        if(glob_id):
            snt = wlk._h.get_obj(wlk._m_id, uuid.UUID(glob_id))
            glob_wlk = snt.walker_ids.get_obj_by_name(wlk.name)
            if(glob_wlk and glob_wlk.code_sig != wlk.code_sig):
                wlk.apply_ir(glob_wlk.code_ir)



from unittest import TestCase
from jaseci.utils.utils import TestCaseHelper
from jaseci.element.super_master import super_master
from jaseci.element.master import master
from jaseci.utils.mem_hook import mem_hook
import jaseci.tests.jac_test_code as jtc


class core_api_test(TestCaseHelper, TestCase):
    """Unit tests for Jac Core APIs"""

    def setUp(self):
        super().setUp()
        self.mast = super_master(h=mem_hook())
        self.mast2 = super_master(h=self.mast._h)
        self.mast.sentinel_register(name='test', code=jtc.basic)

        self.lms = master(h=mem_hook())
        self.lms2 = master(h=self.lms._h)
        self.lms.sentinel_register(name='test', code=jtc.basic)

    def tearDown(self):
        super().tearDown()

    def call(self, mast, pl):
        ret = mast.general_interface_to_api(
            api_name=pl[0], params=pl[1])
        return ret

    def test_sent_global_active(self):
        """Test setting global sentinel"""
        api = ['global_sentinel_set', {'snt': None}]
        self.call(self.mast, api)
        api = ['sentinel_active_global', {}]
        self.call(self.mast2, api)
        api = ['sentinel_list', {'detailed': False}]
        r = self.call(self.mast2, api)
        self.assertEqual(len(r), 0)
        api = ['sentinel_active_get', {'detailed': False}]
        r = self.call(self.mast2, api)
        self.assertIn('jid', r.keys())
        self.assertEqual(r['name'], 'test')
        api = ['sentinel_list', {'detailed': False}]
        r = self.call(self.mast2, api)
        self.assertEqual(len(r), 0)

    def test_sent_global_pull(self):
        """Test setting global sentinel"""
        api = ['global_sentinel_set', {'snt': None}]
        self.call(self.mast, api)
        api = ['sentinel_list', {'detailed': False}]
        r = self.call(self.mast2, api)
        self.assertEqual(len(r), 0)
        api = ['sentinel_pull', {}]
        self.call(self.mast2, api)
        api = ['sentinel_list', {'detailed': False}]
        r = self.call(self.mast2, api)
        self.assertEqual(len(r), 1)
        api = ['sentinel_active_get', {'detailed': False}]
        r = self.call(self.mast2, api)
        self.assertIn('jid', r.keys())
        self.assertEqual(r['name'], 'test')

    def test_global_set_get_delete(self):
        """Test setting global sentinel"""
        api = ['global_get', {'name': 'apple'}]
        r = self.call(self.mast, api)
        self.assertIsNone(r['value'])
        api = ['global_set', {'name': 'apple', 'value': '56'}]
        r = self.call(self.mast, api)
        api = ['global_get', {'name': 'apple'}]
        r = self.call(self.mast2, api)
        self.assertEqual(r['value'], '56')
        api = ['global_delete', {'name': 'apple'}]
        r = self.call(self.mast2, api)
        api = ['global_get', {'name': 'apple'}]
        r = self.call(self.mast, api)
        self.assertIsNone(r['value'])

    def test_master_create(self):
        """Test master create operation"""
        api = ['master_create', {'name': 'yo@gmail.com'}]
        r = self.call(self.lms, api)
        self.assertIn('j_type', r)
        self.assertEqual(r['j_type'], 'master')

    def test_master_create_error_out(self):
        """Test master create operation"""
        api = ['master_create', {'name': 'yo@gmail.com'}]
        r = self.call(self.lms, api)
        api = ['master_create', {'name': 'yo@gmail.com'}]
        r = self.call(self.lms, api)
        self.assertIn('response', r)
        self.assertIn('already exists', r['response'])

    def test_master_create_super_limited(self):
        """Test master create operation"""
        api = ['master_createsuper', {'name': 'yo3@gmail.com'}]
        r = self.call(self.lms, api)
        self.assertIn('response', r)
        self.assertIn('not a valid', r['response'])

    def test_master_create_linked_super_master_create(self):
        """Test master create operation"""
        api = ['master_createsuper', {'name': 'yo3@gmail.com'}]
        r = self.call(self.mast, api)
        self.assertIn('j_type', r)
        self.assertEqual(r['j_type'], 'super_master')

    def test_global_sentinel_set_unset(self):
        api = ['global_sentinel_set', {}]
        r = self.call(self.mast, api)
        self.assertIn('response', r)
        self.assertNotIn('error', r)
        api = ['global_sentinel_unset', {}]
        r = self.call(self.mast, api)
        self.assertIn('response', r)
        self.assertNotIn('error', r)

    def test_global_sentinel_double_unset(self):
        api = ['global_sentinel_set', {}]
        r = self.call(self.mast, api)
        self.assertIn('response', r)
        self.assertNotIn('error', r)
        api = ['global_sentinel_unset', {}]
        r = self.call(self.mast, api)
        self.assertIn('response', r)
        self.assertNotIn('error', r)
        api = ['global_sentinel_unset', {}]
        r = self.call(self.mast, api)
        self.assertIn('response', r)
        self.assertNotIn('error', r)



from jaseci.utils.utils import obj_class_cache, build_class_dict
# from pprint import pformat
from inspect import getdoc
import jaseci


class book():
    def api_cheatsheet(self, root, out=None, str=""):
        if(out is None):
            out = []
        if('leaf' in root.keys()):
            out.append(str+f'& {root["leaf"][1]} &\n')
            return
        for i in root.keys():
            self.api_cheatsheet(root[i], out, str+f'{i} ')
        ret = ''
        for i in out:
            ret += i
        return ret

    def api_spec(self):
        ret = ''
        build_class_dict(jaseci)
        for i in obj_class_cache.keys():
            if(not i.endswith('_api')):
                continue
            ret += f'\\subsection{{{i[:-4]} APIs}}\n\n'
            doc = getdoc(obj_class_cache[i]).replace("\n\n", "\n\\par\n")
            doc = doc.replace('_', '\\_')
            ret += f'{doc}\n\n'
            ret += self.api_call_spec(obj_class_cache[i])
        return ret

    def api_call_spec(self, cls):
        ret = ''
        for i in dir(cls):
            # access = 'master'
            if (i.startswith('api_')):
                api = i[4:].replace('_', ' ')
            elif (i.startswith('admin_api_')):
                # access = 'super'
                api = i[10:].replace('_', ' ')
            elif (i.startswith('public_api_')):
                # access = 'public'
                api = i[11:].replace('_', ' ')
            else:
                continue
            ret += f'\\subsubsection{{{api}}}\n\n'
        return ret



"""
Walker class for Jaseci

Each walker has an id, name, timestamp and it's parent sentinel.
TODO: Perhaps  I should have walker state (context ids) in mem only with
default hooks to save db read/writes
"""

from jaseci.utils.utils import logger
from jaseci.element.element import element
from jaseci.element.obj_mixins import anchored
from jaseci.utils.id_list import id_list
from jaseci.jac.interpreter.walker_interp import walker_interp
from jaseci.jac.ir.jac_code import jac_code
import uuid
import hashlib
import io
import pstats
import cProfile


class walker(element, jac_code, walker_interp, anchored):
    """Walker class for Jaseci"""

    def __init__(self, code_ir=None, *args, **kwargs):
        self.activity_action_ids = id_list(self)
        self.namespaces = []
        self.context = {}
        self.profile = {}
        # Process state
        self.current_node_id = None
        self.next_node_ids = id_list(self)
        self.ignore_node_ids = id_list(self)
        self.destroy_node_ids = id_list(self)
        self.current_step = 0
        self.in_entry_exit = False
        self.step_limit = 10000
        anchored.__init__(self)
        element.__init__(self, *args, **kwargs)
        jac_code.__init__(self, code_ir=code_ir)
        walker_interp.__init__(self)

    @ property
    def current_node(self):
        if(not self.current_node_id):
            return None
        else:
            return self._h.get_obj(self._m_id, uuid.UUID(self.current_node_id))

    @ current_node.setter
    def current_node(self, obj):
        if(obj):
            self.current_node_id = obj.id.urn
        else:
            self.current_node_id = None

    def namespace_keys(self):
        """Return list of md5 keys for namespaces"""
        ret = {}
        for i in self.namespaces:
            ret[i] = hashlib.md5((self._m_id+i).encode()).hexdigest()
        return ret

    def step(self):
        """
        Take single step through program
        if no ast provided, will be generated from code
        """
        if(not self.next_node_ids):
            logger.debug(str(f'Walker {self.name} is disengaged'))
            return False
        if(self.current_step > self.step_limit):
            logger.error(
                str(f'Walker {self.name} walked too many steps ' +
                    f'- {self.step_limit}')
            )
            return False

        self.current_node = self.next_node_ids.pop_first_obj()
        self.run_walker(jac_ast=self._jac_ast)
        if(self.current_step < 200):
            self.log_history('visited', self.current_node.id)
        self.current_step += 1
        self.profile['steps'] = self.current_step
        if (self._stopped == 'skip'):
            self._stopped = None
        if(self.next_node_ids):
            logger.debug(str(f'Step complete, Walker {self.name} next node: ' +
                             f'- {self.next_node_ids.first_obj()}'))
            return self.next_node_ids.first_obj()
        else:
            logger.debug(
                str(f'Final step of walker {self.name} complete' +
                    f'- disengaging from {self.current_node}'))
            for i in self.destroy_node_ids.obj_list():
                i.destroy()
            return True

    def prime(self, start_node, prime_ctx=None):
        """Place walker on node and get ready to step step"""
        self.clear_state()
        self.next_node_ids.add_obj(start_node)
        if (prime_ctx):
            for i in prime_ctx.keys():
                self.context[str(i)] = prime_ctx[i]
        self.profile['steps'] = self.current_step
        logger.debug(str(f'Walker {self.name} primed - {start_node}'))

    def run(self, start_node=None, prime_ctx=None,
            profiling=False):
        """Executes Walker to completion"""
        if(profiling):
            pr = cProfile.Profile()
            pr.enable()

        if(start_node):
            self.prime(start_node, prime_ctx)

        report_ret = {'success': True}
        try:
            while self.step():
                pass
        except Exception as e:
            import traceback as tb
            self.rt_error(f'Internal Exception: {e}', self._cur_jac_ast)
            report_ret['stack_trace'] = tb.format_exc()

        self.save()

        if(not self.report):
            logger.debug(
                str(f'Walker {self.name} did not arrive at report state')
            )

        report_ret['report'] = self.report
        if(len(self.runtime_errors)):
            report_ret['errors'] = self.runtime_errors
            report_ret['success'] = False
        if(profiling):
            pr.disable()
            s = io.StringIO()
            sortby = pstats.SortKey.CUMULATIVE
            ps = pstats.Stats(pr, stream=s).sort_stats(sortby)
            ps.print_stats()
            s = s.getvalue()
            s = 'ncalls'+s.split('ncalls')[-1]
            s = '\n'.join([','.join(line.rstrip().split(None, 5))
                          for line in s.split('\n')])
            self.profile['perf'] = s
            report_ret['profile'] = self.profile

        return report_ret

    def log_history(self, name, value):
        """Helper function for logging history of walker's activities"""
        if (isinstance(value, element)):
            value = {'type': value.j_type, 'id': value.id.urn}
        if (isinstance(value, uuid.UUID)):
            value = value.urn
        if name in self.profile.keys():
            self.profile[name].append(value)
        else:
            self.profile[name] = [value]

    def clear_state(self):
        """Clears walker state after report"""
        self.profile = {}
        self.current_step = 0
        self.next_node_ids.remove_all()
        self.ignore_node_ids.remove_all()
        self.destroy_node_ids.remove_all()
        self.current_node = None
        self.activity_action_ids.destroy_all()
        self.context = {}
        walker_interp.reset(self)

    def destroy(self):
        """
        Destroys self from memory and persistent storage
        """
        for i in self.activity_action_ids.obj_list():
            i.destroy()
        super().destroy()



"""
Logger api as a mixin
"""
from jaseci.api.interface import interface
from jaseci.utils.utils import logger, app_logger
from jaseci.utils.utils import connect_logger_handler
from logging.handlers import HTTPHandler


class logger_api():
    """
    APIs for Jaseci Logging configuration
    """

    @interface.admin_api(cli_args=['host'])
    def logger_http_connect(self, host: str, port: int,
                            url: str, log: str = 'all'):
        """
        Connects internal logging to http(s) (log msgs sent via POSTs)
        Valid log params: {sys, app, all }
        """
        num = 0
        if(log == 'sys' or log == 'all'):
            connect_logger_handler(
                logger, HTTPHandler(host=f'{host}:{port}',
                                    url=url, method='POST'))
            num += 1
        if(log == 'app' or log == 'all'):
            connect_logger_handler(
                app_logger, HTTPHandler(host=f'{host}:{port}',
                                        url=url, method='POST'))
            num += 1
        return [f'{num} http handlers added!']

    @interface.admin_api()
    def logger_http_clear(self, log: str = 'all'):
        """
        Connects internal logging to http(s) (log msgs sent via POSTs)
        Valid log params: {sys, app, all }
        """
        num = 0
        if(log == 'sys' or log == 'all'):
            for i in logger.handlers:
                if(i.__class__.__name__ == 'HTTPHandler'):
                    logger.removeHandler(i)
                    num += 1
        if(log == 'app' or log == 'all'):
            for i in app_logger.handlers:
                if(i.__class__.__name__ == 'HTTPHandler'):
                    app_logger.removeHandler(i)
                    num += 1
        return [f'{num} http handlers removed!']

    @interface.admin_api()
    def logger_list(self):
        """
        Check active loggers
        """
        core = []
        app = []
        for i in logger.handlers:
            core.append(str(type(i)))
        for i in app_logger.handlers:
            app.append(str(type(i)))
        return {'core': core, 'app': app}
